#include <strops.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>

pid_t pid, sid;
int server_sock;
struct sockaddr_un server;
int sockaddr_len;

int main() {
    /* 
     Setup daemon
     Used this for basic idea => https://en.wikipedia.org/wiki/Daemon_(computing)
    */
    pid = fork();
    if (pid == -1) {
        fprintf(stderr, "Couldn't fork\n");
        return 1;
    } else if (pid != 0) {
        printf("Pid of child is %d\n", pid);
        return 0;
    }

    sid = setsid();
    if (sid == -1) {
        return 1;
    }

    if (chdir("/") == -1) {
        return 1;
    }

    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);

    FILE *out_log;
    FILE *err_log;
    /* TODO: switch to mode 'a' later */
    out_log = fopen("/tmp/out_log.txt", "w");
    err_log = fopen("/tmp/err_log.txt", "w");
    if (out_log == NULL || err_log == NULL) {
        return 1;
    }

    fprintf(err_log, "Hello, World!\n");
    
    /* Setup Unix Socket */
    server_sock = socket(AF_UNIX, SOCK_STREAM, 0);
    if (server_sock == -1) {
        return 1;
    }

    server.sun_family = AF_UNIX;
    strcpy(server.sun_path, "/tmp/test_sock");
    unlink(server.sun_path);
    sockaddr_len = strlen(server.sun_path) + sizeof(server.sun_family);

    if (bind(server_sock, (struct sockaddr*)&server, sockaddr_len) == -1) {
        return 1;
    }

    if (listen(server_sock, 3) == -1) {
        return 1;
    }

    fprintf(out_log, "Listening on %s\n", server.sun_path);


    /* Actual program */
    int client_sock;
    int data_length = 0;
    char receive_buf[1000];
    char send_buf[1000];
    memset(receive_buf, 0, sizeof receive_buf);
    memset(send_buf, 0, sizeof send_buf);

    while(1) {
        client_sock = accept(server_sock, NULL, NULL);
        if (client_sock == -1) {
            return 1;
        }
        data_length = recv(client_sock, receive_buf, sizeof receive_buf, 0);
        if (data_length <= 0) {
            strcpy(send_buf, "Failed to receive any data.");
            send(client_sock, send_buf, strlen(send_buf), 0);
            close(client_sock);
            continue;
        }

        if (strstr(send_buf, "shutdown") != NULL) {
            strcpy(send_buf, "Shutting down.");
            send(client_sock, send_buf, strlen(send_buf), 0);
            close(client_sock);
            break;
        }

        close(client_sock);
    }
    
    fclose(out_log);
    fclose(err_log);
    close(server_sock);
    return 0;
}
